# -*- coding: UTF-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this
# file except in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#

''' WebPage模块单元测试
'''

import unittest
from test.util import TestWebView
from qt4w.webcontrols import WebPage, WebElement
import mock

class MyElement(WebElement):
    pass

class testWebPage(unittest.TestCase):
    def _getPage(self):
        testview=TestWebView("TestBrowser")
        return WebPage(testview)

    def test_browser_tye(self):
        page=self._getPage()
        self.assertEqual(page.browser_type,"TestBrowser")

    def test_url(self):
        page=self._getPage()
        self.assertEqual(page.url,"http://www.test.com")

    def test_title(self):
        page = self._getPage()
        self.assertEqual(page.title, "testtitle")

    def test_cookie(self):
        page = self._getPage()
        cookie=page.cookie["Test"]
        self.assertEqual(cookie, "test")

    def test_ready_state(self):
        page = self._getPage()
        self.assertEqual(page.ready_state, "complete")

    def test_wait_for_ready(self):
        page = self._getPage()
        self.assertEqual(page.wait_for_ready(), None)

    def test_wait_for_ready_timeout(self):
        testview = TestWebView("TestBrowser")
        page = WebPage(testview, wait_for_ready=False)
        page._webdriver.set_dict("document.readyState","not Ready")
        self.assertRaises(RuntimeError,page.wait_for_ready, timeout=10)

    def test_upload(self):
        page = self._getPage()
        self.assertEqual(page.upload_file("/root"), "/root upload complete")

    def test_console_log(self):
        page = self._getPage()
        self.assertEqual(page.read_console_log(), "testlog")

    def test_get_element(self):
        page = self._getPage()
        ele=page.get_element("//div[@id='test']")
        self.assertTrue(isinstance(ele,WebElement))

    def test_get_myelement(self):
        page = self._getPage()
        ele = page.get_element("//div[@id='test']",MyElement)
        self.assertTrue(isinstance(ele, MyElement))

    def test_get_elements(self):
        page = self._getPage()
        ele = page.get_elements("//div[@id='test']")
        self.assertEqual(len(ele),3)
        self.assertEqual(ele[0]._locators[0], "(//div[@id='test'])[1]")

if __name__ == '__main__':
    unittest.main()


# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this
# file except in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''WebPage、WebElement接口类
'''


from __future__ import absolute_import,division
import copy
import re
import time
from qt4w.util import Deprecated, TimeoutError, lazy_init, encode_wrap
from qt4w import XPath
from six.moves import xrange
import six

class IWebPage(object):
    '''WebPage接口定义
    '''
    
    @property
    def browser_type(self):
        '''浏览器类型
        '''
        raise NotImplementedError
    
    @property
    def url(self):
        '''页面url
        '''
        raise NotImplementedError
    
    @Deprecated('url')
    @property
    def Url(self):
        '''页面url
        '''
        raise NotImplementedError
    
    @property
    def title(self):
        '''页面标题
        '''
        raise NotImplementedError
    
    @Deprecated('title')
    @property
    def Title(self):
        '''页面标题
        '''
        raise NotImplementedError
    
    @property
    def cookie(self):
        '''页面cookie
        '''
        raise NotImplementedError
    
    @property
    def ready_state(self):
        '''页面状态
        '''
        raise NotImplementedError
    
    @Deprecated('ready_state')
    @property
    def ReadyState(self):
        '''页面状态
        '''
        raise NotImplementedError
    
    @property
    def accessible_object(self):
        '''accessible对象
        '''
        raise NotImplementedError
    
    def activate(self):
        '''激活承载页面的窗口
        '''
        raise NotImplementedError
    
    def close(self):
        '''关闭承载页面的窗口
        '''
        raise NotImplementedError
    
    def release(self):
        '''释放占用的资源
        '''
        raise NotImplementedError
    
    def wait_for_ready(self, timeout=60):
        '''等待页面状态变为ready
        
        :param timeout: 超时时间
        :type timeout:  int
        '''
        raise NotImplementedError
    
    @Deprecated('wait_for_ready')
    def waitForReady(self, timeout=60):
        '''等待页面状态变为ready
        
        :param timeout: 超时时间
        :type timeout:  int
        '''
        raise NotImplementedError
    
    def scroll(self, x, y):
        '''滚动
        
        :param x: 横向滚动的偏移，负值向左，正值向右
        :type x:  int
        :param y: 纵向滚动的偏移，负值向上，正值向下
        :type y:  int
        '''
        raise NotImplementedError
    
    def exec_script(self, script):
        '''在页面中执行JavaScript代码，并返回直接结果
        
        :param script: 要执行的代码
        :type script:  string
        '''
        raise NotImplementedError
    
    @Deprecated('exec_script')
    def execScript(self, script):
        '''在页面中执行JavaScript代码，并返回直接结果
        
        :param script: 要执行的代码
        :type script:  string
        '''
        raise NotImplementedError
    
    def get_element(self, locator):
        '''在页面中查找元素，返回第一个匹配的元素
        
        :param locator: 元素的xpath路径
        :type locator:  string或XPath
        '''
        raise NotImplementedError
    
    @Deprecated('get_element')
    def getElement(self, locator):
        '''在页面中查找元素，返回第一个匹配的元素
        
        :param locator: 元素的xpath路径
        :type locator:  string或XPath
        '''
        raise NotImplementedError
    
    def get_elements(self, locator):
        '''在页面中查找元素，返回包含所有匹配的元素的列表
        
        :param locator: 元素的xpath路径
        :type locator:  string或XPath
        '''
        raise NotImplementedError
    
    @Deprecated('get_elements')
    def getElements(self, locator):
        '''在页面中查找元素，返回包含所有匹配的元素的列表
        
        :param locator: 元素的xpath路径
        :type locator:  string或XPath
        '''
        raise NotImplementedError
    
    def pull_down_refresh(self):
        '''下拉刷新
        '''
        raise NotImplementedError
    
    
    def pull_up_refresh(self):
        '''上拉刷新
        '''
        raise NotImplementedError
    
    def upload_file(self, file_path):
        '''上传文件
        
        :param file_path: 文件路径
        :type  file_path: str
        '''
        raise NotImplementedError
    
    def read_console_log(self, timeout=None):
        '''读取一条console.log输出的日志
        
        :param timeout: 读取日志的超时时间，为None表示不会超时
        :type  timeout: int
        '''
        raise NotImplementedError

    def screenshot(self):
        '''获取page页面截图
        '''
        raise NotImplementedError

class IWebElement(object):
    '''WebElement接口定义
    '''
    @property
    def page(self):
        '''元素所在WebPage
        '''
        raise NotImplementedError
    
    @property
    def attributes(self):
        '''元素的属性集合
        '''
        raise NotImplementedError
    
    @Deprecated('attributes')
    @property
    def Attributes(self):
        '''元素的属性集合
        '''
        raise NotImplementedError
    
    @property
    def styles(self):
        '''元素的样式集合
        '''
        raise NotImplementedError
    
    @property
    def rect(self):
        '''元素位置信息
        '''
        raise NotImplementedError
    
    @property
    def BoundingRect(self):
        '''元素位置信息
        '''
        raise NotImplementedError
    
    @property
    def displayed(self):
        '''元素是否显示
        '''
        raise NotImplementedError
    
    @Deprecated('displayed')
    @property
    def Displayed(self):
        '''元素是否显示
        '''
        raise NotImplementedError
    
    @property
    def visible(self):
        '''元素是否视觉可见
        '''
        raise NotImplementedError
    
    @property
    def focused(self):
        '''元素是否有焦点
        '''
        raise NotImplementedError
    
    @property
    def inner_text(self):
        '''元素所包含的文本
        '''
        raise NotImplementedError
    
    @Deprecated('inner_text')
    @property
    def InnerText(self):
        '''元素所包含的文本
        '''
        raise NotImplementedError
    
    @property
    def inner_html(self):
        '''元素所包含的HTML代码
        '''
        raise NotImplementedError
    
    @Deprecated('inner_html')
    @property
    def InnerHtml(self):
        '''元素所包含的HTML代码
        '''
        raise NotImplementedError
    
    def exist(self):
        '''元素是否存在
        '''
        raise NotImplementedError
    
    def scroll(self, x, y):
        '''滚动元素
        
        :param x: 横向滚动的偏移，负值向左，正值向右
        :type x:  int
        :param y: 纵向滚动的偏移，负值向上，正值向下
        :type y:  int
        '''
        raise NotImplementedError
    
    def get_element(self, locator):
        '''在当前元素的子孙元素中查找元素，返回第一个匹配的元素
        
        :param locator: 元素的xpath路径
        :type locator:  string或XPath
        '''
        raise NotImplementedError
    
    @Deprecated('get_element')
    def getElement(self, locator):
        '''在当前元素的子孙元素中查找元素，返回第一个匹配的元素
        
        :param locator: 元素的xpath路径
        :type locator:  string或XPath
        '''
        raise NotImplementedError
    
    def get_elements(self, locator):
        '''在当前元素的子孙元素中查找元素，返回包含所有匹配的元素的列表
        
        :param locator: 元素的xpath路径
        :type locator:  string或XPath
        '''
        raise NotImplementedError
    
    @Deprecated('get_elements')
    def getElements(self, locator):
        '''在当前元素的子孙元素中查找元素，返回包含所有匹配的元素的列表
        
        :param locator: 元素的xpath路径
        :type locator:  string或XPath
        '''
        raise NotImplementedError
    
    def wait_for_visible(self, timeout=10, interval=0.5):
        '''等待控件可见
        
        :param timeout: 超时时间
        :type timeout:  float
        :param interval:查询间隔时间
        :type interval: float
        '''
        raise NotImplementedError
    
    def wait_for_attribute(self, name, value, timeout=10, interval=0.5):
        '''暂停程序执行，直到当前元素的指定属性变为特定值
        
        :param name:   要等待的属性名
        :type name:    string
        :param value:  要等待的属性值
        :type value:   string
        :param timeout:超时时间
        :type timeout: int或float
        :param interval:重试间隔时间
        :type interval:int或float
        '''
        raise NotImplementedError
    
    def wait_for_style(self, name, value, timeout=10, interval=0.5):
        '''暂停程序执行，直到当前元素的指定样式变为特定值
        
        :param name:   要等待的样式名
        :type name:    string
        :param value:  要等待的样式值
        :type value:   string
        :param timeout:超时时间
        :type timeout: int或float
        :param interval:重试间隔时间
        :type interval:int或float
        '''
        raise NotImplementedError
    
    def wait_for_text(self, value, timeout=10, interval=0.5):
        '''暂停程序执行，直到当前元素的InnerText变为特定值
        
        :param value:   要等待的特定值
        :type value:    string
        :param timeout: 超时时间
        :type timeout:  int或float
        :param interval:重试间隔时间
        :type interval: int或float
        '''
        raise NotImplementedError
    
    def click(self, x_offset=None, y_offset=None):
        '''点击元素，默认点击元素中点
        
        :param x_offset: 距离控件区域左上角的横向偏移。
        :type x_offset:  int或float
        :param y_offset: 距离控件区域左上角的纵向偏移。
        :type y_offset:  int或float
        '''
        raise NotImplementedError
    
    def long_click(self, x_offset=None, y_offset=None, duration=1):
        '''长按元素
        
        :param x_offset: 距离控件区域左上角的横向偏移
        :type x_offset:  int/float
        :param y_offset: 距离控件区域左上角的纵向偏移
        :type y_offset:  int/float
        :param duration: 按住的持续时间
        :type duration:  int/float
        '''
        raise NotImplementedError
    
    def double_click(self, x_offset=None, y_offset=None):
        '''双击元素，默认双击元素中点
        
        :param x_offset: 距离控件区域左上角的横向偏移。
        :type x_offset:  int或float
        :param y_offset: 距离控件区域左上角的纵向偏移。
        :type y_offset:  int或float
        '''
        raise NotImplementedError
    
    def right_click(self, x_offset=None, y_offset=None):
        '''右击元素，默认右击元素中点
        
        :param x_offset: 距离控件区域左上角的横向偏移。
        :type x_offset:  int或float
        :param y_offset: 距离控件区域左上角的纵向偏移。
        :type y_offset:  int或float
        '''
        raise NotImplementedError
    
    def hover(self, x_offset=None, y_offset=None):
        '''鼠标悬停
        
        :param x_offset: 距离控件区域左上角的横向偏移。
        :type x_offset:  int或float
        :param y_offset: 距离控件区域左上角的纵向偏移。
        :type y_offset:  int或float
        '''
        raise NotImplementedError
    
    def drag(self, x, y):
        '''拖放元素到指定位置
        
        :param x: 拖放终点距离起点的横向偏移。
        :type x:  int或float
        :param y: 放终点距离起点的纵向偏移。
        :type y:  int或float
        '''
        raise NotImplementedError
    
    def send_keys(self, keys):
        '''发送按键
        
        :param keys: 发送的按键
        :type keys:  string
        '''
        raise NotImplementedError
    
    @Deprecated('send_keys')
    def sendKeys(self, keys):
        '''发送按键
        
        :param keys: 发送的按键
        :type keys:  string
        '''
        raise NotImplementedError
    
    def set_focus(self):
        '''设为焦点
        '''
        raise NotImplementedError

    def screenshot(self):
        '''设置控件截图
        '''
        raise NotImplementedError

class ControlContainer(object):
    '''控件容器基类
    '''
    ui_control_type = None
    ui_map = {}
    
    def __init__(self):
        self._locator_dict = {}
        self._ui_map = {}
        self._update_ui_map()
        
    def __getitem__(self, key):
        '''获取控件
        '''
        item = copy.copy(self._locator_dict[key])
        cls = item.pop('type')
        root = item.pop('root', None)
        loc = item['locator']
        if isinstance(root, str) and root[0] == '@':
            root = self[root[1:]]
        else:
            root = self

        if issubclass(cls, WebElement):
            return cls(root, loc)
        else:
            return cls(root, key, **item)
            
    @property
    def Controls(self):
        return self
    
    def updateLocator(self, locators):
        '''更新控件定位参数
        
        :type locators: dict
        :param locators: 定位参数，格式是 {'控件名':{'type':控件类, 控件类的参数dict列表}, ...}
        '''
        self._locator_dict.update(locators)
    
    def _update_ui_map(self):
        '''从ui_map中更新控件信息
        '''
        import copy
        cls_list = []
        cls = self.__class__
        while cls != ControlContainer and cls != WebPage and cls != WebElement:
            cls_list.insert(0, cls)  # 基类在前，子类在后
            cls = cls.__base__
        last_cls = None
        for cls in cls_list:
            if cls.ui_map and (not last_cls or cls.ui_map != last_cls.ui_map):
                # 防止某个类未定义ui_map
                ui_map = copy.deepcopy(cls.ui_map)
                self._ui_map.update(ui_map)
            last_cls = cls
    
    def update_ui_map(self, ui_map):
        '''从指定的ui_map中更新控件定义
        '''
        self._ui_map.update(ui_map)
        
    def control(self, name):
        '''获取控件实例
        '''
        from qt4w import XPath
        list_pattern = re.compile(r'^(.+)\[(\d+)\]$')
        if '.' in name:
            # 多层结构方式
            name_list = name.split('.')
            obj = self
            for name in name_list:
                obj = obj.control(name)
            return obj
        
        ret = list_pattern.match(name)
        if ret != None:
            # 索引方式
            name = ret.group(1)
            index = int(ret.group(2))
            return self.control(name)[index]
        
        if not (name in self._ui_map):
            raise NameError("%s没有名为'%s'的子控件！" % (type(self), name))
        ui_control = self._ui_map[name]
        ui_type = self.ui_control_type if self.ui_control_type else WebElement
        if isinstance(ui_control, dict):
            if 'type' in ui_control: ui_type = ui_control['type']
            if not issubclass(ui_type, (WebElement, UIListBase)):
                ui_control.pop('type')
                instance = ui_type(self, name, **ui_control)
                # Not support child control currently
                return instance
        else:
            if not isinstance(ui_control, XPath):
                raise TypeError('控件：%s 类型错误(%s)' % (name, type(XPath)))
            self._ui_map[name] = {'locator': ui_control}
        instance = ui_type(self, self._ui_map[name]['locator'])
        if 'ui_map' in self._ui_map[name]:
            instance._ui_map = self._ui_map[name]['ui_map']
        return instance
    
    def get_metis_view(self):
        '''返回MetisView
        '''
        return MetisView(self)
    
class WebElement(ControlContainer, IWebElement):
    '''IWebElement实现
    '''
    
    def __init__(self, root, locator):
        ControlContainer.__init__(self)
        from qt4w.__init__ import XPath
        from qt4w.util import WebElementAttributes, WebElementStyles
        if isinstance(root, WebPage):
            self._page = root
        else:
            self._page = root.page
        self._webview = self._page._webview
        self._webdriver = self._page._webdriver
        self._root = root
        if isinstance(root, WebPage):
            self._locators = XPath(locator).break_frames()
        elif isinstance(root, FrameElement):
            loc = XPath(locator).break_frames()
            self._locators = root._locators + loc
        else:
            self._locators = root._locators[:-1]
            self._locators += XPath(root._locators[-1] + locator).break_frames()
        self._attrs = WebElementAttributes(self._getattr, self._setattr, self._listattr)
        self._styles = WebElementStyles(self._getstyle)
        
    def __str__(self):
        return '<%s object at 0x%X [xpath=%s]>' % (self.__class__.__name__, id(self), ''.join(self._locators))
    
    def __getitem__(self, key):
        '''获取控件
        '''
        item = self._page._locator_dict[key]
        cls = item['type']
        root = item['root']
        loc = item['locator']
        return cls(self, loc)
    
    def post_init(self):
        '''窗口类自定义的初始化逻辑
        '''
        self._wait_for_exist()
    
    @property
    def page(self):
        '''元素所在WebPage
        '''
        return self._page
    
    @lazy_init
    def _getattr(self, name):
        if name in ["value", "checked", "innerText", "innerHTML"]:
            ret = self._webdriver.get_property(self._locators, name)
        else:
            ret = self._webdriver.get_attribute(self._locators, name)
        return ret
    
    @lazy_init
    def _setattr(self, name, value):
        self.highlight()
        if name in ["value", "checked", "innerText", "innerHTML"]:
            ret = self._webdriver.set_property(self._locators, name, value)
        else:
            ret = self._webdriver.set_attribute(self._locators, name, value)
        return ret

    def _listattr(self):
        return ['id', 'name', 'class']
    
    @lazy_init
    def _getstyle(self, name):
        ret = self._webdriver.get_style(self._locators, name.lower())
        return ret
    
    def _wait_for_exist(self, timeout=10, interval=0.5):
        '''等待元素出现
        
        :param timeout:  超时时间，单位为秒
        :type timeout:   int/float
        :param interval: 重试时间间隔，单位为秒
        :type interval:  int/float
        '''
        time0 = time.time()
        while time.time() - time0 < timeout:
            if self.exist(): return
            time.sleep(interval)
        self._webdriver.get_element(self._locators)
        
    @property
    def attributes(self):
        '''元素的属性集合
        '''
        return self._attrs
    
    @property
    def styles(self):
        '''元素的样式集合
        '''
        return self._styles
    
    @property
    @lazy_init
    def rect(self):
        '''元素相对于WebView的位置信息
        '''
        return self._webdriver.get_elem_rect(self._locators, False)


    def screenshot(self):
        '''设置控件截图
        '''
        if not self.visible:
            self._webdriver.scroll_to_visible(self._locators)
        page=self.page.screenshot()
        left=self.rect[0]
        top=self.rect[1]
        width=self.rect[2]
        heigth=self.rect[3]
        return page.crop(box=(left,top,left+width,top+heigth))

    @property
    def BoundingRect(self):
        '''元素位置信息
        '''
        from qt4w.util import Rect
        return Rect(self.rect)
    
    @property
    @lazy_init
    def displayed(self):
        '''元素是否显示在页面中（Web层可见，肉眼不一定可见）
        '''
        if self.styles['display'] == 'none': return False
        if self.styles['visibility'] == 'hidden': return False  # 只有为hidden的时候才认为不可见
        self_rect = self.rect
        if self_rect[2] == 0 or self_rect[3] == 0: return False
        return True
    
    @property
    def visible(self):
        '''元素是否视觉可见
        '''
        if not self.displayed: return False
        self_rect = self.rect
        root_rect = self._webview.visible_rect
        root_rect = [0, 0, root_rect[2], root_rect[3]]
        
        if self_rect[0] >= root_rect[0] and \
            self_rect[0] + self_rect[2] <= root_rect[0] + root_rect[2] and \
            self_rect[1] >= root_rect[1] and \
            self_rect[1] + self_rect[3] <= root_rect[1] + root_rect[3]:
            return True
        return False
    
    @property
    @lazy_init
    def focused(self):
        '''元素是否有焦点
        '''
        return self._webdriver.is_elem_focused(self._locators)

    @property
    @lazy_init
    @encode_wrap
    def inner_text(self):
        '''元素所包含的文本
        '''
        return self._attrs['innerText']
    
    @inner_text.setter
    @lazy_init
    def inner_text(self, text):
        '''设置元素所包含的文本
        '''
        self._attrs['innerText'] = text
    
    @property
    @encode_wrap
    def inner_html(self):
        '''元素所包含的HTML代码
        '''
        return self._attrs['innerHTML']
    
    @inner_html.setter
    def inner_html(self, html):
        '''设置html代码
        '''
        self._attrs['innerHTML'] = html
        
    def exist(self):
        '''元素是否存在
        '''
        from qt4w.util import ControlNotFoundError
        try:
            self._webdriver.get_element(self._locators)
            return True
        except ControlNotFoundError:
            return False
    
    @lazy_init
    def highlight(self):
        '''高亮
        '''
        self._webdriver.highlight(self._locators)
        time.sleep(0.5) # wait for highlight
        
    def scroll(self, x, y):
        '''滚动元素
        
        :param x: 横向滚动的偏移，负值向左，正值向右
        :type x:  int
        :param y: 纵向滚动的偏移，负值向上，正值向下
        :type y:  int
        '''
        raise NotImplementedError
    
    def get_element(self, locator, elem_cls=None):
        '''在当前元素的子孙元素中查找元素，返回第一个匹配的元素
        
        :param locator: 元素的xpath路径
        :type locator:  string或XPath
        :paran elem_cls: 返回的元素类型
        :type elem_cls:  class
        '''
        if elem_cls == None: elem_cls = WebElement
        return elem_cls(self, locator)
    
    def get_elements(self, locator, elem_cls=None):
        '''在当前元素的子孙元素中查找元素，返回包含所有匹配的元素的列表
        
        :param locator: 元素的xpath路径
        :type locator:  string或XPath
        :paran elem_cls: 返回的元素类型
        :type elem_cls:  class
        '''
        locator = self._locators[-1] + locator
        return self._page.get_elements(locator, elem_cls)
    
    def wait_for_visible(self, timeout=10, interval=0.5):
        '''等待控件可见
        
        :param timeout: 超时时间
        :type timeout:  float
        :param interval:查询间隔时间
        :type interval: float
        '''
        time0 = time.time()
        while time.time() - time0 < timeout:
            if self.visible: return
            time.sleep(interval)
        else:
            raise RuntimeError('元素：%s 在%s秒内不可见' % (''.join(self._locators), timeout))
        
        
    def wait_for_attribute(self, name, value, timeout=10, interval=0.5):
        '''暂停程序执行，直到当前元素的指定属性变为特定值
        
        :param name:   要等待的属性名
        :type name:    string
        :param value:  要等待的属性值
        :type value:   string
        :param timeout:超时时间
        :type timeout: int或float
        :param interval:重试间隔时间
        :type interval:int或float
        '''
        time0 = time.time()
        while time.time() - time0 < timeout:
            real_value = self.attributes[name]
            real_value = real_value.replace('"', '')
            if real_value == value:return
            time.sleep(interval)
        raise TimeoutError('等待控件属性%s超时：期望值："%s"，当前值："%s"' % (name, value, self.inner_text))
    
    def wait_for_style(self, name, value, timeout=10, interval=0.5):
        '''暂停程序执行，直到当前元素的指定样式变为特定值
        
        :param name:   要等待的样式名
        :type name:    string
        :param value:  要等待的样式值
        :type value:   string
        :param timeout:超时时间
        :type timeout: int或float
        :param interval:重试间隔时间
        :type interval:int或float
        '''
        raise NotImplementedError
    
    def wait_for_text(self, text, timeout=10, interval=0.5):
        '''暂停程序执行，直到当前元素的InnerText变为特定值
        
        :param text:   要等待的特定值
        :type text:    string
        :param timeout: 超时时间
        :type timeout:  int或float
        :param interval:重试间隔时间
        :type interval: int或float
        '''
        time0 = time.time()
        while time.time() - time0 < timeout:
            if self.inner_text == text:return
            time.sleep(interval)
        raise TimeoutError('等待控件文本超时：期望值："%s"，当前值："%s"' % (text, self.inner_text))
    
    def wait_for_value(self, value, timeout=10, interval=0.5):
        '''暂停程序执行，直到当前元素的InnerText变为特定值
        
        :param value:   要等待的特定值
        :type value:    string
        :param timeout: 超时时间
        :type timeout:  int或float
        :param interval:重试间隔时间
        :type interval: int或float
        '''
        
    
    def _pre_click(self, x_offset=None, y_offset=None, highlight=True):
        '''点击前的处理
        '''
        # 等待控件可见
        timeout = 5
        time0 = time.time()
        while time.time() - time0 < timeout:
            if self.displayed: break
            time.sleep(0.1)
        else:
            raise TimeoutError('控件：%s 在%d秒内不可见' % (self, timeout))
        
        self._webdriver.scroll_to_visible(self._locators)

        rect = self.rect
        time0 = time.time()
        while time.time() - time0 < 2:
            time.sleep(0.5)  # 等待元素位置发生变化
            new_rect = self.rect
            if new_rect != rect: break
            self._webdriver.scroll_to_visible(self._locators) # 避免某些情况下滑动失败
            
        if highlight: self.highlight()
        rect = self.rect  # 此时坐标可能发生变化，需要重新获取
        outer_rect = self._webview.visible_rect
        outer_rect = [0, 0, outer_rect[2], outer_rect[3]]  # 换算成以WebView左上角为原点的坐标
        # 计算交集，避免计算出的坐标为窗口外面
        left = max(rect[0], outer_rect[0])
        top = max(rect[1], outer_rect[1])
        right = min(rect[0] + rect[2], outer_rect[0] + outer_rect[2])
        bottom = min(rect[1] + rect[3], outer_rect[1] + outer_rect[3])
        rect = (left, top, right - left, bottom - top)

        x_offset = rect[0] + (rect[2] / 2 if not x_offset else x_offset)
        y_offset = rect[1] + (rect[3] / 2 if not y_offset else y_offset)
        return x_offset, y_offset
    
    def click(self, x_offset=None, y_offset=None, highlight=True):
        '''点击元素，默认点击元素中点
        
        :param x_offset: 距离控件区域左上角的横向偏移。
        :type x_offset:  int或float
        :param y_offset: 距离控件区域左上角的纵向偏移。
        :type y_offset:  int或float
        :param highlight: 是否高亮元素
        :type  highlight: bool
        '''
        x_offset, y_offset = self._pre_click(x_offset, y_offset, highlight)
        self._webview.click(x_offset, y_offset)
    
    def long_click(self, x_offset=None, y_offset=None, duration=1):
        '''长按元素
        
        :param x_offset: 距离控件区域左上角的横向偏移
        :type x_offset:  int/float
        :param y_offset: 距离控件区域左上角的纵向偏移
        :type y_offset:  int/float
        :param duration: 按住的持续时间
        :type duration:  int/float
        '''
        x_offset, y_offset = self._pre_click(x_offset, y_offset)
        self._webview.long_click(x_offset, y_offset, duration)
        
    def double_click(self, x_offset=None, y_offset=None):
        '''双击元素，默认双击元素中点
        
        :param x_offset: 距离控件区域左上角的横向偏移。
        :type x_offset:  int或float
        :param y_offset: 距离控件区域左上角的纵向偏移。
        :type y_offset:  int或float
        '''
        x_offset, y_offset = self._pre_click(x_offset, y_offset)
        self._webview.double_click(x_offset, y_offset)
        
    def right_click(self, x_offset=None, y_offset=None):
        '''右击元素，默认右击元素中点
        
        :param x_offset: 距离控件区域左上角的横向偏移。
        :type x_offset:  int或float
        :param y_offset: 距离控件区域左上角的纵向偏移。
        :type y_offset:  int或float
        '''
        x_offset, y_offset = self._pre_click(x_offset, y_offset)
        self._webview.right_click(x_offset, y_offset)
    
    def hover(self, x_offset=None, y_offset=None):
        '''鼠标悬停
        
        :param x_offset: 距离控件区域左上角的横向偏移。
        :type x_offset:  int或float
        :param y_offset: 距离控件区域左上角的纵向偏移。
        :type y_offset:  int或float
        '''
        x_offset, y_offset = self._pre_click(x_offset, y_offset)
        self._webview.hover(x_offset, y_offset)
    
    def drag(self, x, y):
        '''拖放元素到指定位置
        
        :param x: 拖放终点距离起点的横向偏移。
        :type x:  int或float
        :param y: 放终点距离起点的纵向偏移。
        :type y:  int或float
        '''
        screen_size = self._webdriver.get_screen_size()
        if abs(x) > screen_size[0]: x = screen_size[0] * abs(x) / x
        if abs(y) > screen_size[1]: y = screen_size[1] * abs(y) / y
        
        rect = self.rect
        # 计算与屏幕的交集
        if rect[0] < 0: 
            rect[2] += rect[0]
            rect[0] = 0
        if rect[1] < 0: 
            rect[3] += rect[1]
            rect[1] = 0
        if rect[0] + rect[2] > screen_size[0]:
            rect[2] = screen_size[0] - rect[0]
        if rect[1] + rect[3] > screen_size[1]:
            rect[3] = screen_size[1] - rect[1]
        mid_x = rect[0] + rect[2] / 2
        mid_y = rect[1] + rect[3] / 2

        self._webdriver.drag_element(self._locators, mid_x - x / 2, mid_y - y / 2, mid_x + x / 2, mid_y + y / 2)
        time.sleep(1)  # 等待滚动完成
        
    def send_keys(self, keys):
        '''发送按键
        
        :param keys: 发送的按键
        :type keys:  string
        '''
        self.click()
        self._webview.send_keys(keys)
    
    def set_focus(self):
        '''设为焦点
        '''
        raise NotImplementedError  

class WebPage(ControlContainer, IWebPage):
    '''IWebPage接口实现
    '''
    ui_control_type = WebElement
    WebElement.ui_control_type = WebElement  # 不能直接在WebElement中赋值
            
    def __init__(self, webview_or_webpage, locator=None, wait_for_ready=True):
        ControlContainer.__init__(self)
        if locator is None: locator = []
        if isinstance(webview_or_webpage, WebPage):
            self._webview = webview_or_webpage._webview
            self._locator = webview_or_webpage._locator
            self._webdriver = webview_or_webpage._webdriver
        else:
            self._webview = webview_or_webpage
            self._locator = locator
            try:
                self._webdriver = self._webview.webdriver_class(self._webview)
            except NotImplementedError:
                self._webdriver = self._webview  # remove later
        if wait_for_ready: self.wait_for_ready()
        
    def __str__(self):
        return '<%s object at 0x%X [title=%s url=%s]>' % (self.__class__.__name__, id(self), self.title, self.url)
    
    @property
    def browser_type(self):
        '''浏览器类型
        '''
        return self._webview.browser_type
        
    @property
    def url(self):
        '''页面url
        '''
        return self.exec_script('location.href')
    
    @url.setter
    def url(self, new_url):
        '''跳转到指定url
        
        :param new_url: 要跳转到的url
        :type  new_url: str
        '''
        self.exec_script('location.href="%s"' % new_url)
        
    @property
    def title(self):
        '''页面标题
        '''
        return self.exec_script('document.title')
    
    @property
    def cookie(self):
        '''页面cookie
        '''
        return self.exec_script('document.cookie')
    
    @property
    def ready_state(self):
        '''页面状态
        '''
        return self.exec_script('document.readyState')
    
    def activate(self):
        '''激活承载页面的窗口
        '''
        raise NotImplementedError
    
    def close(self):
        '''关闭承载页面的窗口
        '''
        self.exec_script('close()')
    
    def release(self):
        '''释放占用的资源
        '''
        raise NotImplementedError
    
    def wait_for_ready(self, timeout=60):
        '''等待页面状态变为ready
        
        :param timeout: 超时时间
        :type timeout:  int
        '''
        time0 = time.time()
        while time.time() - time0 < timeout:
            if self.ready_state == 'complete': return
            time.sleep(0.5)
        else:
            raise RuntimeError('页面未在%d秒内加载完成' % timeout)
    
    def scroll(self, x, y):
        '''滚动
        
        :param x: 横向滚动的偏移，负值向左，正值向右
        :type x:  int
        :param y: 纵向滚动的偏移，负值向上，正值向下
        :type y:  int
        '''
        self.exec_script('scrollBy(%s, %s);' % (x, y))
    
    @encode_wrap
    def exec_script(self, script):
        '''在页面中执行JavaScript代码，并返回直接结果
        
        :param script: 要执行的代码
        :type script:  string
        '''
        return self._webdriver.eval_script(self._locator, script)
    
    def get_element(self, locator, elem_cls=None):
        '''在页面中查找元素，返回第一个匹配的元素
        
        :param locator: 元素在当前页面的xpath路径
        :type locator:  string或XPath
        :paran elem_cls: 返回的元素类型
        :type elem_cls:  class
        '''
        if elem_cls == None: elem_cls = WebElement
        return elem_cls(self, locator)
    
    def get_elements(self, locator, elem_cls=None):
        '''在页面中查找元素，返回包含所有匹配的元素的列表
        
        :param locator: 元素的xpath路径
        :type locator:  string或XPath
        :paran elem_cls: 返回的元素类型
        :type elem_cls:  class
        '''
        from qt4w.util import LazyDict
        locators = self._locator[:]
        locators.append(locator)
        elem_count = self._webdriver.get_element_count(locators)
        def get_elem(index):
            loc = '(%s)[%d]' % (locator, index + 1)
            return self.get_element(loc, elem_cls)

        return LazyDict(get_elem, lister=lambda:xrange(elem_count))
    
    def pull_down_refresh(self):
        '''下拉刷新
        '''
        self.exec_script('scroll(0, 0);')  # 滑动到顶部 
        rect = self._webview.visible_rect
        x1 = x2 = rect[0] + rect[2] / 2
        y1 = rect[1] + rect[3] / 4
        y2 = rect[1] + rect[3] * 3 / 4
        self._webview.drag(x1, y1, x2, y2)  # 从纵坐标1/4处滑倒3/4处

    def pull_up_refresh(self):
        '''上拉刷新 
        '''
        self.exec_script('scroll(0,document.body.scrollHeight);')  # 滑动到底部 
        rect = self._webview.visible_rect
        x1 = x2 = rect[0] + rect[2] / 2
        y1 = rect[1] + rect[3] * 3 / 4
        y2 = rect[1] + rect[3] / 4
        self._webview.drag(x1, y1, x2, y2)  # 从纵坐标3/4处滑倒1/4处
    
    def upload_file(self, file_path):
        '''上传文件
        
        :param file_path: 文件路径
        :type  file_path: str
        '''
        return self._webview.upload_file(file_path)
    
    def read_console_log(self, timeout=None):
        '''读取一条console.log输出的日志
        
        :param timeout: 读取日志的超时时间，为None表示不会超时
        :type  timeout: int
        '''
        time0 = time.time()
        while timeout == None or time.time() - time0 < timeout:
            result = self._webdriver.read_console_log(self._locator)
            if result: return result
            time.sleep(0.1)
        else:
            raise TimeoutError('Read console log timeout')

    def screenshot(self):
        '''获取page页面截图
        '''
        self._webview.screenshot()
        
class FrameElement(WebElement):
    '''frame/iframe元素
    '''
    def __init__(self, root, locator):
        loc = XPath(locator).break_frames()[-1]
        if not ('/iframe' in loc or '/frame' in loc):
            raise ValueError("Not a Frame nor IFrame: '%s'" % locator)
        super(FrameElement, self).__init__(root, locator)
        self._frame_page = None
    
    def __str__(self):
        return '<%s object at 0x%X [locator=%s]>' % (self.__class__.__name__, id(self), ''.join(self._locators))
    
    @property
    def framepage(self):
        '''frame中包含的WebPage
        '''
        if self._frame_page is None:
            webview = self._webview
            self._frame_page = WebPage(webview, self._locators)
            self._frame_page.updateLocator(self._page._locator_dict)
        return self._frame_page
    
    @property
    def FramePage(self):
        '''
        '''
        return self.framepage

class InputElement(WebElement):
    '''input元素
    '''
    @property
    def value(self):
        '''当前的value
        '''
        return self._webdriver.get_property(self._locators, 'value')
    
    @value.setter
    def value(self, val):
        '''设置新的value
        
        :param val: 新设置的value
        :type  val: string
        '''
        self._webdriver.set_property(self._locators, 'value', str(val))
        self._webdriver.fire_event(self._locators, 'input')
        self._webdriver.fire_event(self._locators, 'change')
        
class SelectElement(WebElement):
    '''select元素
    '''
    
    @property
    def options(self):
        '''返回所有选项列表
        '''
        option_list = []
        for it in self.get_elements('/option'):
            option_list.append(it.inner_text.strip())
        return option_list
    
    @property
    def selection(self):
        '''当前选择项
        '''
        return self._webdriver.get_property(self._locators, 'selectedIndex')
    
    @selection.setter
    def selection(self, option):
        '''选择某个选项
        
        :param option: 要选中的选项索引或显示的内容
        :type  option: int/string
        '''
        if isinstance(option, (str, six.text_type)):
            # change to option index
            option = self.options.index(option)

        self._webdriver.set_property(self._locators, 'selectedIndex', option)
        self._webdriver.fire_event(self._locators, 'change')
        
        
class UIListBase(object):
    '''List控件基类
    '''
    ui_control_type = None
    
    def __init__(self, root, locator):
        self._root = root
        self._locator = locator
        self._ui_map = {}
        self._elements = []
    
    def _get_elements(self):
        '''获取元素列表
        '''
        if self._elements: return self._elements
        result = []
        elements = list(self._root.get_elements(self._locator, self.ui_control_type))
        for elem in elements:
            elem._ui_map.update(self._ui_map)
            elem._parent = self._root  # elem._root是WebPage对象
            result.append(elem)
        self._elements = result
        return result
    
    def __len__(self):
        '''list长度
        '''
        return len(self._get_elements())
    
    def __iter__(self):
        '''迭代器
        '''
        for elem in self._get_elements():
            yield elem
    
    def __getitem__(self, index):
        '''索引方式访问
        '''
        if not isinstance(index, (int, six.integer_types)):
            raise TypeError('索引值必须为整数：%r' % index)
        if index < 0: index += len(self)
        if index >= len(self): raise IndexError('索引越界，数组长度为：%d，当前索引值为： %d' % (len(self), index))
        elem = self._get_elements()[index]
        return elem
    
    def filter(self, condition):
        '''根据条件过滤，找到满足条件的项即返回，如果找不到则抛出异常
        
        :param condition: 过滤条件
        :type condition:  dict
        '''
        if not condition: raise RuntimeError('过滤条件不能为空')
        pattern = re.compile(r'^(\w+)\[(.+)\]$')
        for elem in self._get_elements():
            equal_flag = True
            for key in condition:
                if not '.' in key:
                    raise ValueError('过滤条件错误：%s' % key)
                control_name, attr_name = key.split('.')
                child_elem = elem.control(control_name)
                index = ''
                ret = pattern.match(attr_name)
                if ret != None:
                    attr_name = ret.group(1)
                    index = ret.group(2)[1:-1]
                if not hasattr(child_elem, attr_name):
                    raise RuntimeError('控件：%s 没有属性：%s' % (child_elem, attr_name))
                value = getattr(child_elem, attr_name)
                if index: value = value.__getitem__(index)
                if value != condition[key]:
                    equal_flag = False
                    break
            if equal_flag: return elem
        raise RuntimeError('未找到满足条件：%s 的子控件' % condition)

def ui_list(control_cls):
    '''列表类型
    '''
    return type('%sList' % control_cls.__name__, (UIListBase,), {'ui_control_type': control_cls})

class MetisView(object):
    '''实现IMetisView接口
    '''
    
    def __init__(self, page_or_elem):
        if isinstance(page_or_elem, WebPage):
            self._page = page_or_elem
            self._elem = None
        else:
            self._page = page_or_elem.page
            self._elem = page_or_elem
        self._webview = self._page._webview
        
    @property
    def rect(self):
        '''元素相对坐标(x, y, w, h)
        '''
        if self._elem == None:
            rect = self._webview.rect
            return (0, 0, rect[2], rect[3])
        else:
            return self._elem.rect
            
    @property
    def os_type(self):
        '''系统类型，例如"android"，"ios"，"pc"
        '''
        return 'web'

    def screenshot(self):
        '''当前容器的区域截图
        :return: PIL.image
        '''
        image = self._webview.screenshot()
        if self._elem != None:
            image = image.crop(self.rect)
        return image
    
    def _get_position(self, offset_x=None, offset_y=None):
        '''
        '''
        if offset_x == None: offset_x = 0.5
        if offset_y == None: offset_y = 0.5
        rect = self.rect
        x = rect[0] + int(rect[2] * offset_x)
        y = rect[1] + int(rect[3] * offset_y)
        return x, y
        
    def click(self, offset_x=None, offset_y=None):
        '''点击
        :param offset_x: 相对于该控件的坐标offset_x，百分比( 0 -> 1 )，不传入则默认该控件的中央
        :type offset_x: float|None
        :param offset_y: 相对于该控件的坐标offset_y，百分比( 0 -> 1 )，不传入则默认该控件的中央
        :type offset_y: float|None
        '''
        x, y = self._get_position(offset_x, offset_y)
        self._webview.click(x, y)

    def send_keys(self, text):
        '''发送可见字符按键
        
        :param text: 要输入的文本
        :type  text: string
        '''
        self._webview.send_keys(text)

    def double_click(self, offset_x=None, offset_y=None):
        '''双击
        '''
        x, y = self._get_position(offset_x, offset_y)
        self._webview.double_click(x, y)

    def long_click(self, offset_x=None, offset_y=None):
        '''长按
        '''
        x, y = self._get_position(offset_x, offset_y)
        self._webview.long_click(x, y)
            
if __name__ == '__main__':
    pass
    



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this
# file except in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''IBrowser接口定义
'''

from __future__ import absolute_import
import random
from qt4w.util import logger

class IBrowser(object):
    '''浏览器接口类
    '''
    
    def open_url(self, url, page_cls=None):
        '''打开一个url，返回page_cls类的实例
        
        :param url: 要打开页面的url
        :type url:  string
        :param page_cls: 要返回的具体WebPage类,为None表示返回WebPage实例
        :type page_cls: Class
        '''
        raise NotImplementedError
    
    def find_by_url(self, url, page_cls=None):
        '''在当前打开的页面中查找指定url,返回page_cls类的实例，如果未找到，返回None
        
        :param url: 要查找的页面url
        :type url:  string
        :param page_cls: 要返回的具体WebPage类,为None表示返回WebPage实例
        :type page_cls: Class
        '''
        raise NotImplementedError
    
class Browser(IBrowser):
    '''对外的浏览器类
    '''
    browser_dict = {}  # 存储浏览器类型与浏览器类的对应关系
    
    def __init__(self, browser_name=None):
        '''创建具体的Browser实例
        
        :param browser_name: 要创建的浏览器类型
        :type browser_name:
        '''
        self._browser_name = browser_name
        self._browser = self._get_browser_cls()
    
    def _get_browser_cls(self):
        '''获取浏览器类
        '''
        if not self._browser_name:
            # 随机选择浏览器
            self._browser_name = random.choice(self.browser_dict.keys())
        browser_cls_path = self.browser_dict.get(self._browser_name)
            
        if not browser_cls_path:
            raise TypeError('Browser %s is not registered' % self._browser_name)
        logger.info('[Browser] Current browser type is %s' % self._browser_name)
        
        module = __import__('.'.join(browser_cls_path.split('.')[:-1]))
        for item in browser_cls_path.split('.')[1:]:
            module = getattr(module, item)
        return module()
    
    @staticmethod
    def register_browser(browser_name, browser_cls_path):
        '''注册浏览器
        
        :param browser_name:     浏览器名称
        :type browser_name:      string
        :param browser_cls_path: 浏览器类路径
        :type browser_cls_path:  string
        '''
        Browser.browser_dict[browser_name] = browser_cls_path
        
    def open_url(self, url, page_cls=None):
        '''打开一个url，返回page_cls类的实例
        
        :param url: 要打开页面的url
        :type url:  string
        :param page_cls: 要返回的具体WebPage类,为None表示返回WebPage实例
        :type page_cls: Class
        '''
        return self._browser.open_url(url, page_cls)
    
    def find_by_url(self, url, page_cls=None, timeout=10):
        '''在当前打开的页面中查找指定url,返回WebPage实例，如果未找到，返回None
        
        :param url: 要查找的页面url
        :type url:  string
        :param page_cls: 要返回的具体WebPage类,为None表示返回WebPage实例
        :type page_cls: Class
        :param timeout: 查找超时时间，单位：秒
        :type timeout: int/float
        '''
        page = self._browser.find_by_url(url, page_cls, timeout)
        if page == None: raise RuntimeError('Can\'t find page %s in browser %s' % (url, self._browser_name))
        return page
        
if __name__ == '__main__':
    pass
    



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this
# file except in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''IWebDriver接口
'''

import json
import six

from qt4w.util import logger, general_encode, unicode_decode, ControlNotFoundError


class IWebDriver(object):
    '''IWebDriver接口
    '''
    def __init__(self, webview):
        '''构造函数
        
        :param webview: 对应的WebView实例
        :type webview:  object
        '''
        raise NotImplementedError
    
    @property
    def ready_state(self):
        '''页面状态
        '''
        raise NotImplementedError
    
    def eval_script(self, frame_xpaths, script):
        '''在指定frame中执行JavaScript，并返回执行结果
        
        :param frame_xpaths: frame元素的XPATH路径，如果是顶层页面，怎传入“[]”
        :type frame_xpaths:  list
        :param script:       要执行的JavaScript语句
        :type script:        string
        '''
        raise NotImplementedError
    
    def get_attribute(self, elem_xpaths, attr_name):
        '''获取元素属性
        
        :param elem_xpaths: 元素的XPATH路径
        :type elem_xpaths:  list
        :param attr_name:   属性名
        :type attr_name:    string
        '''
        raise NotImplementedError
    
    def set_attribute(self, elem_xpaths, attr_name, value):
        '''设置元素属性
        
        :param elem_xpaths: 元素的XPATH路径
        :type elem_xpaths:  list
        :param attr_name:   属性名
        :type attr_name:    string
        :param value:       新的值
        :type value:        string
        '''
        raise NotImplementedError
        
    def get_elem_rect(self, elem_xpaths, rav=True):
        '''获取元素在页面中的坐标
        
        :param elem_xpaths: 元素的XPATH路径
        :type elem_xpaths:  list
        :param rav:         是否是相对于当前frame
        :type rav:          bool
        '''
        raise NotImplementedError
    
    def get_property(self, elem_xpaths, prop_name):
        '''获取元素的特定值，例如：node.innerHTML
        
        :param elem_xpaths: 元素的XPATH路径
        :type elem_xpaths:  list
        :param prop_name:   property名
        :type prop_name:    string
        '''
        raise NotImplementedError
    
    def set_property(self, elem_xpaths, prop_name, value):
        '''设置元素的特定值，例如：node.innerHTML
        
        :param elem_xpaths: 元素的XPATH路径
        :type elem_xpaths:  list
        :param prop_name:   property名
        :type prop_name:    string
        :param value:       新的值
        :type value:        string
        '''
        raise NotImplementedError
    
    def get_style(self, elem_xpaths, style_name):
        '''获取元素的某一样式值
        
        :param elem_xpaths: 元素的XPATH路径
        :type elem_xpaths:  list
        :param style_name:  样式名称
        :type style_name:   string
        '''
        raise NotImplementedError
    
    def highlight(self, elem_xpaths):
        '''使元素高亮
        
        :param elem_xpaths: 元素的XPATH路径
        :type elem_xpaths:  list
        '''
        raise NotImplementedError

class WebDriverBase(IWebDriver):
    '''WebDriver基类
    '''
    driver_script = r'''
    if (typeof window.qt4w_hook_console == 'undefined') window.qt4w_hook_console = true;
    window['qt4w_driver_lib'] = {
    getScale : function(){
        return 1;
    },
    
    getScreenSize: function(){
        var result = new Array();
        result.push(screen.availWidth);
        result.push(screen.availHeight);
        return result.toString();
    },
    
    selectNodes : function(xpath){
        var oResult = document.evaluate(xpath, document, null, XPathResult.ANY_TYPE, null);
        var aNodes = new Array();  
        if (oResult != null) {  
            var oElement = oResult.iterateNext();  
            while (oElement) {  
                aNodes.push(oElement);  
                oElement = oResult.iterateNext();  
            }  
        }
        return aNodes;
    },
    
    selectNode : function(xpath){
        var nodes = this.selectNodes(xpath);
        if(nodes.length == 0) throw new Error('Find element '+xpath+' failed');
        else if(nodes.length > 1) throw new Error('Find '+nodes.length+' elements match '+xpath);
        return nodes[0];
    },
    
    getElementZoom: function(node){
        var scale = 1;
        while(node != null && node != document.documentElement){
            var zoom = parseFloat(window.getComputedStyle(node, null).zoom);
            if(zoom != 1){
                scale *= zoom;
            }
            node = node.parentNode;
        }
        return scale;
    },
    
    getElemRect: function(node){
        var result = new Array();
        var rect = node.getBoundingClientRect();
        var scale = this.getScale();
        scale *= this.getElementZoom(node);
        var left = rect.left;
        var top = rect.top;
        var width = rect.width;
        var height = rect.height;
        if (typeof width == 'undefined') width = rect.right - rect.left;
        if (typeof height == 'undefined') height = rect.bottom - rect.top;
        result.push(left * scale);
        result.push(top * scale);
        result.push(width * scale);
        result.push(height * scale);
        return result.toString();
    },
    
    initHighlightDiv: function(){
        this.bd0 = document.createElement("div");
        this.bd1 = document.createElement("div");
        this.bd2 = document.createElement("div");
        this.bd3 = document.createElement("div");
    },
    
    showDiv: function(cnt){
        if(cnt % 2 != 0){
            document.body.appendChild(this.bd0);
            document.body.appendChild(this.bd1);
            document.body.appendChild(this.bd2);
            document.body.appendChild(this.bd3);
        } else {
            document.body.removeChild(this.bd0);
            document.body.removeChild(this.bd1);
            document.body.removeChild(this.bd2);
            document.body.removeChild(this.bd3);
        }
        if (cnt){
            cnt--;
            if(window.console) console.log('show' + cnt);
            setTimeout("qt4w_driver_lib.showDiv(" + cnt + ")", 100);
        }
    },
    
    highlight: function(node){
        var rect = node.getBoundingClientRect();
        var left= rect.left;
        var top = rect.top;
        var width = node.offsetWidth;
        var height = node.offsetHeight;
        if(window.console) console.log(left+','+top+','+width+','+height);
        
        this.bd0.setAttribute("style", "border:solid 1px red;"
            + "left:" + (left) + "px;top:" + (top) + "px;z-index:32767;"
            + "width:" + (width) + "px;height:0px;position:fixed;");
        this.bd1.setAttribute("style", "border:solid 1px red;"
            + "left:" + (left) + "px;top:" + (top) + "px;z-index:32767;"
            + "width:0px;height:" + (height) + "px;position:fixed;");
        this.bd2.setAttribute("style", "border:solid 1px red;"
            + "left:" + (left+width) + "px;top:" + (top) + "px;z-index:32767;"
            + "width:0px;height:" + (height) + "px;position:fixed;");
        this.bd3.setAttribute("style", "border:solid 1px red;"
            + "left:" + (left) + "px;top:" + (top+height) + "px;z-index:32767;"
            + "width:" + (width) + "px;height:0px;position:fixed;");
        //console.log('style'+this.bd0.getAttribute('style')+'');
        this.showDiv(3);
    },
    
    scrollToVisible: function(node){
        if(node.scrollIntoViewIfNeeded){
            node.scrollIntoViewIfNeeded();
        }else if(node.scrollIntoView){
            node.scrollIntoView();
        }
    },
    
    logData: [],
    hookConsole: function(){
        var self = this;
        if(window.console && console.log && console.log.apply && window.JSON){
            var hookedConsoleFunc = function(){
                var timeStr = new Date().toLocaleString();
                for(var i=1;i<arguments.length;i++){
                    var data = arguments[i];
                    if(data instanceof Array){
                        data = JSON.stringify(data);
                    }else if(data instanceof Object){
                        var jsonData = {};
                        for(var key in data){
                            jsonData[key] = data[key];
                            if(jsonData[key]) jsonData[key] = jsonData[key].toString();
                        }
                        data = JSON.stringify(jsonData);
                    }
                    self.logData.push('[' + timeStr + '][console.' + arguments[0].name + '] ' + data);
                }
                var args = Array.prototype.slice.call(arguments);
                args.splice(0, 1);
                return arguments[0].apply(this, args);
            }

            var hookFunction = function (funcName) {
                if (!console[funcName]) return;
                var origFunc = console[funcName];
                console[funcName] = function(){
                    var args = Array.prototype.slice.call(arguments);
                    args.splice(0, 0, origFunc);
                    return hookedConsoleFunc.apply(this, args);
                }
            }

            hookFunction('log');
            hookFunction('dir');
            hookFunction('info');
            hookFunction('warn');
            hookFunction('error');
        }
    },
    
    readLogData: function (count) {
        if (count < 0) count = this.logData.length;
        return this.logData.splice(0, count);
    }
    
    };
    qt4w_driver_lib.initHighlightDiv();
    if (window.qt4w_hook_console) qt4w_driver_lib.hookConsole();
    '''
    
    def __init__(self, webview):
        '''构造函数
        
        :param webview: 对应的WebView实例
        :type webview:  object
        '''
        self._webview = webview
        
    @staticmethod
    def create_driver(webview):
        '''根据webview类型创建对应的WebDriver实例
        
        :param webview: WebView实例
        :type webview:  object
        '''
        web_driver_path = webview.web_driver
        module = __import__(web_driver_path)
        for item in web_driver_path.split('.')[1:]:
            module = getattr(module, item)
        web_driver_cls = getattr(module, 'WebDriver')
        return web_driver_cls(webview)
    
    def _my_encode(self, text):
        '''对于中文，统一处理成unicode编码
        如“中国”，变成“\u4e2d\u56fd”
        '''
        text = unicode_decode(text)
        return general_encode(text.encode('raw_unicode_escape'))

    def _my_decode(self, text):
        return text.decode('raw_unicode_escape')
    
    def _xpath_encode(self, xpath):
        xpath = xpath.replace('\'', '"')
        return xpath
        # return self._my_encode(xpath)
    
    def _xpaths_encode(self, xpath_list):
        for i in range(len(xpath_list)):
            xpath_list[i] = self._xpath_encode(xpath_list[i])
    
    def _break_xpaths(self, elem_xpaths):
        '''将xpath数组分隔成frame_xpaths和elem_xpath
        '''
        if len(elem_xpaths) < 1: raise ValueError('xpath is []')
        frame_xpaths = elem_xpaths[:-1]
        self._xpaths_encode(frame_xpaths)
        elem_xpath = elem_xpaths[-1]
        elem_xpath = self._xpath_encode(elem_xpath)
        return frame_xpaths, elem_xpath
    
    def _get_dom_tree(self, frame_xpaths):
        '''获取DOM树
        '''
        try:
            from tuia.env import run_env, EnumEnvType
            if run_env != EnumEnvType.Lab: return ''
        except ImportError:
            pass
        
        result = '\nCurrent DOM Tree：\n'
        dom_tree = self.eval_script(frame_xpaths, 'document.documentElement.outerHTML;')
        if isinstance(dom_tree, six.text_type): dom_tree = dom_tree.encode('utf8')
        result += dom_tree
        return result
    
    def eval_script(self, frame_xpaths, script):
        '''在指定frame中执行JavaScript，并返回执行结果（该实现需要处理js基础库未注入情况的处理）
        
        :param frame_xpaths: frame元素的XPATH路径，如果是顶层页面，怎传入“[]”
        :type frame_xpaths:  list
        :param script:       要执行的JavaScript语句
        :type script:        string
        '''
        from qt4w.util import JavaScriptError, ControlNotFoundError, ControlAmbiguousError
        try:
            return self._webview.eval_script(frame_xpaths, script)
        except JavaScriptError as e:
            err_msg = e.message
            err_msg = err_msg.split('\n')[0]  # 错误堆栈信息可能会有影响
            err_msg = general_encode(err_msg)
            if ('ReferenceError' in err_msg or 'TypeError' in err_msg) and 'qt4w_driver_lib' in err_msg:
                # ie8 TypeError ie9 ReferenceError
                # 注入js基础库
                self.eval_script(e.frame, self.driver_script)
                return self.eval_script(frame_xpaths, script)
            elif 'Find element' in err_msg and 'failed' in err_msg:
                # 未找到元素
                raise ControlNotFoundError(err_msg)
            elif 'Find' in err_msg and 'elements match' in err_msg:
                err_msg += self._get_dom_tree(frame_xpaths)
                raise ControlAmbiguousError(err_msg)
            else:
                raise e
    
    def get_ready_state(self, frame_xpaths):
        '''获取页面状态
        '''
        return self.eval_script(frame_xpaths, 'document.readyState')
    
    def get_screen_size(self):
        '''获取屏幕大小
        '''
        result = self.eval_script([], 'qt4w_driver_lib.getScreenSize();')
        width, height = result.split(',')
        return int(width), int(height)
    
    def _get_frame_info(self, frame_xpaths):
        '''获取frame属性信息
        '''
        frame_xpath = frame_xpaths[-1].replace('\'', '"')
        js = r'''
        var frame_node = qt4w_driver_lib.selectNode('%s');
        var url = frame_node.getAttribute('src');
        if (url[0] == '/') {
            url = location.protocol + '//' + location.host + url;
        }
        (frame_node.getAttribute('name') || frame_node.getAttribute('id') || '') + ',' + url;
        ''' % (frame_xpath)
        # 优先使用name，没有name使用id
        result = self.eval_script(frame_xpaths[:-1], js)
        pos = result.find(',')
        if pos < 0: raise RuntimeError('Get frame info failed')
        return result[:pos], result[pos + 1:]
    
    def get_element(self, elem_xpaths):
        '''获取控件，如果控件不存在抛出ControlNotFoundError
        
        :param elem_xpaths: 元素xpath路径
        :type elem_xpaths:  list
        '''
        frame_xpaths, elem_xpath = self._break_xpaths(elem_xpaths)
        js = '''qt4w_driver_lib.selectNode('%s');''' % elem_xpath
        try:
            return self.eval_script(frame_xpaths, js)
        except ControlNotFoundError as e:
            logger.exception(self._get_dom_tree(frame_xpaths))
            err_msg = e.message
            raise ControlNotFoundError(err_msg)
        
    def get_element_count(self, elem_xpaths):
        '''获取满足elem_xpaths的元素个数
        
        :param elem_xpaths: 元素的XPATH路径
        :type elem_xpaths:  list
        '''
        frame_xpaths, elem_xpath = self._break_xpaths(elem_xpaths)
        js = '''qt4w_driver_lib.selectNodes('%s').length;''' % elem_xpath
        return int(self.eval_script(frame_xpaths, js))
    
    def get_attribute(self, elem_xpaths, attr_name):
        '''获取元素属性
        
        :param elem_xpaths: 元素的XPATH路径
        :type elem_xpaths:  list
        :param attr_name:   属性名
        :type attr_name:    string
        '''
        frame_xpaths, elem_xpath = self._break_xpaths(elem_xpaths)
        js = '''
            var node = qt4w_driver_lib.selectNode('%s');
            node.getAttribute('%s');
        ''' % (elem_xpath, attr_name)
        result = self.eval_script(frame_xpaths, js)
        if result == 'undefined': result = None
        return result
    
    def set_attribute(self, elem_xpaths, attr_name, value):
        '''设置元素属性
        
        :param elem_xpaths: 元素的XPATH路径
        :type elem_xpaths:  list
        :param attr_name:   属性名
        :type attr_name:    string
        :param value:       新的值
        :type value:        string
        '''
        frame_xpaths, elem_xpath = self._break_xpaths(elem_xpaths)

        if type(value) == str:
            value = self._my_encode(value)
            value = '\'' + value + '\''
        js = '''
            qt4w_driver_lib.node = qt4w_driver_lib.selectNodes('%s')[0];
            if(qt4w_driver_lib.node == undefined) throw('find %s failed');
            qt4w_driver_lib.result = qt4w_driver_lib.node.setAttribute('%s', %s);
        ''' % (elem_xpath, elem_xpath, attr_name, value)
        return self.eval_script(frame_xpaths, js)
        
    def get_elem_rect(self, elem_xpaths, rav=True):
        '''获取元素在页面中的坐标
        
        :param elem_xpaths: 元素的XPATH路径
        :type elem_xpaths:  list
        :param rav:         是否是相对于当前frame
        :type rav:          bool
        '''
        frame_xpaths, elem_xpath = self._break_xpaths(elem_xpaths)
        
        result = self._get_elem_rect(frame_xpaths, elem_xpath)
        if not rav and frame_xpaths:
            while frame_xpaths:
                result1 = self._get_elem_rect(frame_xpaths[:-1], frame_xpaths[-1])
                for i in range(2):
                    result[i] += result1[i]
                frame_xpaths = frame_xpaths[:-1]
        return result
    
    def _get_elem_rect(self, frame_xpaths, elem_xpath):
        '''获取元素在当前frame中的相对坐标
        '''
        js = '''
            var node = qt4w_driver_lib.selectNode('%s');
            qt4w_driver_lib.result = qt4w_driver_lib.getElemRect(node);
        ''' % elem_xpath  
        result = self.eval_script(frame_xpaths, js)
        result = result.replace('"', '')
        result = result.split(',')
        for i in range(len(result)):
            result[i] = float(result[i])
        return result
    
    def get_property(self, elem_xpaths, prop_name):
        '''获取元素的特定值，例如：node.innerHTML
        
        :param elem_xpaths: 元素的XPATH路径
        :type elem_xpaths:  list
        :param prop_name:   property名
        :type prop_name:    string
        '''
        frame_xpaths, elem_xpath = self._break_xpaths(elem_xpaths)

        js = '''
            var node = qt4w_driver_lib.selectNode('%s');
            node.%s;
        ''' % (elem_xpath, prop_name)
        return self.eval_script(frame_xpaths, js)
    
    def set_property(self, elem_xpaths, prop_name, value):
        '''设置元素的特定值，例如：node.innerHTML
        
        :param elem_xpaths: 元素的XPATH路径
        :type elem_xpaths:  list
        :param prop_name:   property名
        :type prop_name:    string
        :param value:       新的值
        :type value:        string
        '''
        frame_xpaths, elem_xpath = self._break_xpaths(elem_xpaths)

        if value is None:
            value = ''
        if type(value) == str:
            value = self._my_encode(value)
            value = '\'' + value + '\''
        js = '''
            var node = qt4w_driver_lib.selectNode('%s');
            node.%s = %s;
        ''' % (elem_xpath, prop_name, value)
        return self.eval_script(frame_xpaths, js)
    
    def get_style(self, elem_xpaths, style_name):
        '''获取元素的某一样式值
        
        :param elem_xpaths: 元素的XPATH路径
        :type elem_xpaths:  list
        :param style_name:  样式名称
        :type style_name:   string
        '''
        frame_xpaths, elem_xpath = self._break_xpaths(elem_xpaths)

        js = r'''
            var node = qt4w_driver_lib.selectNode('%s');
            window.getComputedStyle(node, null).getPropertyValue('%s');
        ''' % (elem_xpath, style_name)
        return self.eval_script(frame_xpaths, js)
    
    def is_elem_focused(self, elem_xpaths):
        '''是否是当前有焦点元素
        
        :param elem_xpaths:
        :type elem_xpaths:
        '''
        frame_xpaths, elem_xpath = self._break_xpaths(elem_xpaths)
        js = r'''
            var node = qt4w_driver_lib.selectNode('%s');
            document.activeElement == node;
        ''' % (elem_xpath)
        return 'true' in self.eval_script(frame_xpaths, js)
    
    def scroll_to_visible(self, elem_xpaths):
        '''将元素滚动到可见区域
        
        :param elem_xpaths: 元素的XPATH路径
        :type elem_xpaths:  list
        '''
        frame_xpaths, elem_xpath = self._break_xpaths(elem_xpaths)
        js = r'''qt4w_driver_lib.scrollToVisible(qt4w_driver_lib.selectNode('%s'));''' % elem_xpath
        self.eval_script(frame_xpaths, js)
    
    def highlight(self, elem_xpaths):
        '''使元素高亮
        
        :param elem_xpaths: 元素的XPATH路径
        :type elem_xpaths:  list
        '''
        self.scroll_to_visible(elem_xpaths)
        frame_xpaths, elem_xpath = self._break_xpaths(elem_xpaths)
        
        js = '''qt4w_driver_lib.highlight(qt4w_driver_lib.selectNode('%s'));''' % elem_xpath
        self.eval_script(frame_xpaths, js)
        return True
    
    def drag_element(self, elem_xpaths, from_x, from_y, to_x, to_y):
        '''拖拽元素
        
        :param elem_xpaths: 元素的XPATH路径
        :type elem_xpaths:  list
        :param from_x:      起点横坐标，相对于WebView左上角
        :type from_x:       int/float
        :param from_y:      起点纵坐标，相对于WebView左上角
        :type from_y:       int/float
        :param to_x:        终点横坐标，相对于WebView左上角
        :type to_x:         int/float
        :param to_y:        终点纵坐标，相对于WebView左上角
        :type to_y:         int/float
        '''
        self.scroll_to_visible(elem_xpaths)
        frame_xpaths, elem_xpath = self._break_xpaths(elem_xpaths)
        
        js = '''
window['qt4w_driver_lib']['createTouch'] = function(element, point, id) {
    if (document.createTouch)
        return document.createTouch(window, element, id, point.x, point.y, point.x, point.y);

    return {
        view: window,
        target: element,
        identifier: id || 0,
        pageX: point.x,
        pageY: point.y,
        clientX: point.x,
        clientY: point.y
    };
};

window['qt4w_driver_lib']['createTouchList'] = function(touches) {
    if (document.createTouchList)
        return document.createTouchList(touches);

    if (Array.isArray(touches))
        return touches;

    return [touches];
};

window['qt4w_driver_lib']['fireTouchStartEvent'] = function(element, x, y){
    var startEvent = document.createEvent("Event");
    startEvent.initEvent('touchstart', true, true);
    startEvent.touches = this.createTouchList(this.createTouch(element, { x: x, y: y}, 1));
    startEvent.changedTouches = startEvent.touches;
    element.dispatchEvent(startEvent);
};

window['qt4w_driver_lib']['fireTouchMoveEvent'] = function(element, x, y){
    var moveEvent = document.createEvent("Event");
    moveEvent.initEvent('touchmove', true, true);
    moveEvent.touches = this.createTouchList(this.createTouch(element, { x: x, y: y}, 1));
    moveEvent.changedTouches = moveEvent.touches;
    element.dispatchEvent(moveEvent);
};

window['qt4w_driver_lib']['fireTouchEndEvent'] = function(element, x, y){
    var endEvent = document.createEvent("Event");
    endEvent.initEvent('touchend', true, true);
    endEvent.touches = this.createTouchList(this.createTouch(element, { x: x, y: y}, 1));
    endEvent.changedTouches = endEvent.touches;
    element.dispatchEvent(endEvent);
};

window['qt4w_driver_lib']['fireDragEvent'] = function(element, x1, y1, x2, y2){
    this.fireTouchStartEvent(element, x1, y1);
    var self = this;
    setTimeout(function(){self.fireTouchMoveEvent(element, (x1+x2)/2, (y1+y2)/2);}, 200);
    setTimeout(function(){self.fireTouchEndEvent(element, x2, y2);}, 400);
};

qt4w_driver_lib.fireDragEvent(qt4w_driver_lib.selectNode('%s'), %s, %s, %s, %s);
        ''' % (elem_xpath, from_x, from_y, to_x, to_y)
        self.eval_script(frame_xpaths, js)
    
    def fire_event(self, elem_xpaths, type):
        '''触发事件
        
        :param elem_xpaths: 要触发事件的元素XPATH路径
        :type  elem_xpaths: list
        :param type:        事件类型
        :type  type:        string
        '''
        frame_xpaths, elem_xpath = self._break_xpaths(elem_xpaths)
        
        js = '''
var node = qt4w_driver_lib.selectNode('%s');
var evt = document.createEvent("Events");
evt.initEvent('%s', true, false);
node.dispatchEvent(evt);
        ''' % (elem_xpath, type)
        self.eval_script(frame_xpaths, js)
    
    def read_console_log(self, frame_xpaths, count=1):
        '''读取指定条数的日志
        
        :param frame_xpaths: 当前页面的XPATH路径
        :type  frame_xpaths: list
        :param count:        要读取的日志条数，默认为1
        :type  count:        int
        :return:             读取到的日志
        '''
        js = '''JSON.stringify(qt4w_driver_lib.readLogData(%d));''' % count
        result = json.loads(self.eval_script(frame_xpaths, js))
        if count == 1 and len(result) == 1:
            return result[0]
        return result

        
if __name__ == '__main__':
    pass
    



# -*- coding: UTF-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this
# file except in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#

''' browser模块单元测试
'''


import unittest
import mock
from qt4w.browser.browser import Browser

class TestBrowser(unittest.TestCase):


    def test_register_browser(self):
        Browser.register_browser("testBrowser","test.util.FakeBrowser")
        self.assertEqual(Browser.browser_dict["testBrowser"],"test.util.FakeBrowser")

    def test_openurl(self):
        Browser.register_browser("testBrowser", "test.util.FakeBrowser")
        self.assertEqual("http://www.test.com", Browser("testBrowser").open_url("http://www.test.com"))



if __name__ == '__main__':
    unittest.main()


# -*- coding: UTF-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this
# file except in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#

''' WebPage模块单元测试
'''

import unittest
import mock
from qt4w import XPath

class TestXPath(unittest.TestCase):

    def test_node(self):
        xp=XPath('//div[@id="google_shimpl"]')
        self.assertEqual("iv",xp.Nodetest);
        xp = XPath('//div[@id="google_shimpl"]//div[text()="test"]')
        self.assertEqual("div", xp.Nodetest);


    def test_break_step(self):
        xp = XPath('//div[@id="google_shimpl"]//div[text()="test"]')
        steps=xp.break_steps()
        self.assertEqual('//div[@id="google_shimpl"]', steps[0]);
        self.assertEqual('//div[text()="test"]', steps[1]);

    def test_break_frame(self):
         xp = XPath('//div[@id="google_shimpl"]//div[text()="test"]')
         steps = xp.break_frames()
         self.assertEqual('//div[@id="google_shimpl"]//div[text()="test"]', steps[0]);
         xp = XPath('//iframe[@id="google_shimpl"]//div[text()="test"]')
         steps = xp.break_frames()
         self.assertEqual('//iframe[@id="google_shimpl"]', steps[0]);

if __name__ == '__main__':
    unittest.main()


# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this
# file except in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#

'''Webkit WebDrvier实现
'''


from __future__ import absolute_import
from qt4w.webdriver.webdriver import WebDriverBase

class WebkitWebDriver(WebDriverBase):
    '''Webkit WebDrvier
    '''
    
    driver_script = WebDriverBase.driver_script + r'''
    window.Notification = undefined; // disable Notification
    window['qt4w_driver_lib']['getScale'] = function(){return window.devicePixelRatio;};
    window['qt4w_driver_lib']['getElemRect'] = function(node) {
        var result = new Array();
        var rect = node.getBoundingClientRect();
        var scale = this.getScale();
        scale *= this.getElementZoom(node);
        var left = rect.left;
        var top = rect.top;
        if (document.documentElement.scrollWidth > document.documentElement.clientWidth) {
            // 页面未适配终端
            if (window.visualViewport) {
                // above Chrome 61
                left -= window.visualViewport.offsetLeft;
                top -= window.visualViewport.offsetTop;
            } else {
                if (window.scrollX && document.documentElement.getBoundingClientRect().left == 0) {
                    // getBoundingClientRect return fix position
                    left -= window.scrollX;
                }
                if (window.scrollY && document.documentElement.getBoundingClientRect().top == 0) {
                    top -= window.scrollY;
                }
            }
        }
        
        result.push(left * scale);
        result.push(top * scale);
        result.push(rect.width * scale);
        result.push(rect.height * scale);
        return result.toString();
    }
    document.addEventListener('click', function(event){console.log('[ClickListener](' + event.clientX + ', ' + event.clientY + ')');}, true);
    '''



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this
# file except in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''QT4W公共库
'''

from __future__ import absolute_import, print_function
import sys, os
import logging as logger
import six

class QT4WRuntimeError(RuntimeError):
    '''QT4W运行时错误
    '''
    
    @property
    def message(self):
        '''解决python3上没有message属性的问题
        '''
        return self.args[0]
    
class JavaScriptError(RuntimeError):
    '''执行JavaScript报错
    '''
    def __init__(self, frame, err_msg):
        super(JavaScriptError, self).__init__(err_msg)
        self._frame = frame
        self._err_msg = err_msg
        
    @property
    def frame(self):
        '''发生JS异常的frame
        '''
        return self._frame
    
    @property
    def message(self):
        return self._err_msg
    
    def __str__(self):
        return '%s%s' % (self.frame, self._err_msg)

class ControlNotFoundError(QT4WRuntimeError):
    '''Web元素未找到
    '''
    pass
    

class ControlAmbiguousError(QT4WRuntimeError):
    '''找到多个控件
    '''
    pass

class TimeoutError(QT4WRuntimeError):
    '''超时错误
    '''
    pass
    
class LazyDict(object):
    '''类字典容器，本身不存储数据，只在需要时调用相应函数实现读写操作'''

    def __init__(self, getter, setter=None, lister=None):
        self._getter = getter
        self._setter = setter
        self._lister = lister

    def __getitem__(self, key):
        return self._getter(key)

    def __setitem__(self, key, value):
        if not self._setter:
            raise Exception("Item cannot be set.")
        return self._setter(key, value)

    def __delitem__(self):
        raise Exception("Item cannot be deleted.")

    def __iter__(self):
        if not self._lister:
            raise Exception("Items cannot be listed.")
        keys = self._lister()
        for key in keys:
            yield self.__getitem__(key)        

    def __len__(self):
        return len(self._lister())

    def __str__(self):
        keys = self._lister()
        ret = "{"
        for key in keys:
            ret += "%s: %s, " % (repr(key), repr(self[key]))
        return ret[:-2] + "}"

class WebElementAttributes(LazyDict):
    '''供WebElement的Attributes属性使用的类字典容器'''
    def __delitem__(self):
        raise Exception("Attribute cannot be deleted.")

class WebElementStyles(LazyDict):
    '''供WebElement的Styles属性使用的类字典容器'''
    def __setitem__(self, key, value):
        raise Exception("Style cannot be set.")

    def __delitem__(self):
        raise Exception("Style cannot be deleted.")

def general_encode(s):
    '''字符串通用编码处理
    python2 => utf8
    python3 => unicode
    '''
    if six.PY2 and isinstance(s, (unicode, )):
        s = s.encode('utf8')
    elif six.PY3 and isinstance(s, (bytes,)):
        s = s.decode('utf8')
    return s

def unicode_decode(s):
    '''将字符串解码为unicode编码
    '''
    if six.PY2 or isinstance(s, (bytes,)):
        s = s.decode('utf8')
    return s
    
def encode_wrap(func):
    '''处理函数返回值编码
    '''
    def wrap_func(*args, **kwargs):
        ret = func(*args, **kwargs)
        return general_encode(ret)
    return wrap_func

class Deprecated(object):
    '''废弃函数包装
    '''
    def __init__(self, new_func):
        self._new_func = new_func
    
    def __call__(self, func):
        def wrap_func(this, *args, **kwargs):
            # print 'call', func
            frame = sys._getframe(1)
            code = frame.f_code
            file_name = os.path.split(code.co_filename)[-1]
            print >> sys.stderr, '[Warning] method [%s] is deprecated, called in [%s:%s], pls use [%s] instead' % (func.__name__, file_name, code.co_name, self._new_func)
            return getattr(this, self._new_func)(*args, **kwargs)
        
        if func.__class__.__name__ == 'function': 
            return wrap_func
        elif isinstance(func, property): 
            def prop_fget(fget):
                def _wrap_fget(this):
                    frame = sys._getframe(1)
                    code = frame.f_code
                    file_name = os.path.split(code.co_filename)[-1]
                    print >> sys.stderr, '[Warning] property [%s] is deprecated, called in [%s:%s], pls use [%s] instead' % (fget.__name__, file_name, code.co_name, self._new_func)
                    return getattr(this, self._new_func)
                return _wrap_fget
            return property(prop_fget(func.fget), func.fset, func.fdel)
        else:
            raise NotImplementedError(func.__class__.__name__)

def lazy_init(func):
    '''懒初始化
    '''
    def _wrap_func(self, *args, **kwargs):
        if not hasattr(self, '__inited') or self.__inited != True:
            self.post_init()
            self.__inited = True
        return func(self, *args, **kwargs)
    return _wrap_func

class Rect(object):
    '''控件坐标区域
    '''
    def __init__(self, rect):
        self._rect = rect  # [left, top, width, height]
    
    def __getitem__(self, index):
        if not isinstance(index, (int, six.integer_types)):
            raise IndexError('index must be integer')
        if index >= 0 and index < 4:
            return self._rect[index]
        else:
            raise IndexError('index %d out of range' % index)
    
    @property
    def left(self):
        '''左上角横坐标
        '''
        return self._rect[0]
    
    @Deprecated('left')
    @property
    def Left(self):
        '''左上角横坐标
        '''
        raise NotImplementedError
    
    @property
    def top(self):
        '''左上角纵坐标
        '''
        return self._rect[1]
    
    @Deprecated('top')
    @property
    def Top(self):
        '''左上角纵坐标
        '''
        raise NotImplementedError
    
    @property
    def width(self):
        '''宽度
        '''
        return self._rect[2]
    
    @Deprecated('width')
    @property
    def Width(self):
        '''宽度
        '''
        raise NotImplementedError
    
    @property
    def height(self):
        '''高度
        '''
        return self._rect[3]
    
    @Deprecated('height')
    @property
    def Height(self):
        '''高度
        '''
        raise NotImplementedError

class Frame(object):
    '''frame element
    '''
    
    def __init__(self, _id, name, url):
        self._id = _id
        self._name = name
        self._url = url
        self._children = []
    
    @property
    def id(self):
        return self._id
    
    @property
    def name(self):
        return self._name
    
    @property
    def url(self):
        return self._url
    
    def add_child(self, frame):
        '''添加子frame
        '''
        self._children.append(frame)
    
    def find_child_frame(self, name, url):
        '''查找子frame
        '''
        for child in self._children:
            if (name and name == child.name) or (url and url == child.url):
                return child
        return None
    
class FrameSelector(object):
    '''Frame定位
    '''
    
    def __init__(self, webdriver, root_frame):
        '''
        :param webdriver: WebDriver实例
        :param root_frame: 顶层Frame对象
        '''
        self._webdriver = webdriver
        self._root_frame = root_frame
        
    def get_frame_by_xpath(self, frame_xpaths):
        '''根据XPath对象查找frame
        
        :param frame_xpaths: frame的xpath数组
        :type frame_xpaths: list
        :return: Frame object
        '''
        if not frame_xpaths: return self._root_frame
        frame = self._root_frame
        for i in range(len(frame_xpaths)):
            name, url = self._webdriver._get_frame_info(frame_xpaths[:i + 1])
            frame = frame.find_child_frame(name, url)
            if frame == None:
                # frame not found
                return None
        return frame

if __name__ == '__main__':
    pass



# -*- coding: UTF-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this
# file except in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#

''' Webelement模块单元测试
'''

from __future__ import absolute_import
import unittest
import mock
from qt4w.webview.webview  import IWebView
from qt4w.webcontrols import WebElement ,ui_list,UIListBase
from test.util import TestPage,TestElement
from past.builtins import cmp


class TestWebElement(unittest.TestCase):
    def get_control(self):
        test = TestPage()
        return test.control("test")

    def test_page(self):
        element= self.get_control()
        self.assertTrue(isinstance(element.page,TestPage))
        self.assertEqual("testbrowser",element.page.browser_type)

    def test_controls(self):
        test = TestPage()
        element=test.control("test")
        self.assertTrue(isinstance(element,WebElement))
        self.assertEqual("//div[@id='test']",element._locators[0])
        close = test.control("test.close")
        self.assertTrue(isinstance(close, WebElement))
        self.assertEqual( "//div[@id='test']//div[@class='close']",close._locators[0])


    def test_rect(self):
        element = self.get_control()
        self.assertEqual(0,cmp(element.rect,[0,1,2,3]))

    def test_Boundrect(self):
        element = self.get_control()
        self.assertEqual(0, element.BoundingRect.left )
        self.assertEqual(1, element.BoundingRect.top)
        self.assertEqual(2, element.BoundingRect.width)
        self.assertEqual(3, element.BoundingRect.height)

    def test_dispaly(self):
        test = TestPage()
        element = test.control("test")
        #验证display为None
        style={"display":"none"}
        test.get_webview().set_attr_dict(style)
        self.assertFalse(element.displayed)
        # 验证display不为None
        style = {"display": "True","visibility":"display"}
        test.get_webview().set_attr_dict(style)
        self.assertTrue(element.displayed)
        # 验证visibility为hidden
        style = {"display": "True", "visibility": "hidden"}
        test.get_webview().set_attr_dict(style)
        self.assertFalse(element.displayed)
        # 验证width或者heigth为0
        test.get_webview().set_rect([0,0,0,1])
        self.assertFalse(element.displayed)
        test.get_webview().set_rect([0, 0, 1, 0])
        self.assertFalse(element.displayed)

    def test_visible(self):
        test = TestPage()
        #diaplay 不显示
        style = {"display": "none"}
        test.get_webview().set_attr_dict(style)
        element = test.control("test")
        self.assertFalse(element.visible)
        # 控件正常显示
        style = {"display": "true","visibility":"display"}
        test.get_webview().set_attr_dict(style)
        self.assertTrue(element.visible)
        #控件显示不全
        test.get_webview().set_rect([0,1,2,5])
        self.assertFalse(element.visible)

    def test_innertext(self):
        test = TestPage()
        attr = {"innerText": "innerTextTest"}
        test.get_webview().set_attr_dict(attr)
        element = test.control("test")
        self.assertEqual("innerTextTest",element.inner_text)

    def test_set_innertext(self):
        test = TestPage()
        attr = {"innerText": "innerTextTest"}
        test.get_webview().set_attr_dict(attr)
        element = test.control("test")
        element.inner_text= "setInnertexttest"
        self.assertEqual("setInnertexttest",element.inner_text)

    def test_innerhtml(self):
        test = TestPage()
        attr = {"innerHTML": "innerhtmlTest"}
        test.get_webview().set_attr_dict(attr)
        element = test.control("test")
        self.assertEqual("innerhtmlTest",element.inner_html)

    def test_set_innerhtml(self):
        test = TestPage()
        attr = {"innerHTML": "innerHtmlTest"}
        test.get_webview().set_attr_dict(attr)
        element = test.control("test")
        element.inner_html= "setInnerhtmltest"
        self.assertEqual("setInnerhtmltest",element.inner_html)

    def test_exist(self):
        element=self.get_control()
        self.assertTrue(element.exist())

    def test_pre_click(self):
        test = TestPage()
        style = {"display": "true", "visibility": "display"}
        test.get_webview().set_attr_dict(style)
        element = test.control("test")
        xoff,yoff=element._pre_click()
        self.assertEqual(1,xoff)
        self.assertEqual(2.5,yoff)
        xoff, yoff = element._pre_click(x_offset=1, y_offset=1)
        self.assertEqual(1, xoff)
        self.assertEqual(2, yoff)

class TestUIListBase(unittest.TestCase):
    def get_uilist(self):
        test=TestElement()
        return ui_list(TestElement)(test,"//div[id@='test']")

    def test_len(self):
        uilist=self.get_uilist()
        self.assertEqual(3, len(uilist))

    def test_getelements(self):
        uilist = self.get_uilist()
        self.assertTrue(isinstance(uilist._get_elements()[0],TestElement))
        self.assertEqual(3, len(uilist._get_elements()))

    def test_fikter_condition_null(self):
        uilist = self.get_uilist()
        self.assertRaises( RuntimeError,uilist.filter,'')

    def test_filter_attrbute_null(self):
        uilist = self.get_uilist()
        self.assertRaises(RuntimeError,uilist.filter,{"test.value":"test"})

    def test_filter_atrr(self):
        uilist = self.get_uilist()
        self.assertRaises(RuntimeError,uilist.filter, {"test.attributes['value']":"value"})
        self.assertTrue(isinstance(uilist.filter({"test.attributes['value']": "test"}),TestElement))

if __name__ == '__main__':
    unittest.main()


# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this
# file except in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''Web自动化测试基础库
'''



class XPath(str):
    '''表示XPath'''
    def __init__(self, obj):
        str.__init__(self)
        self._obj = obj
        self._censored = None

    def __str__(self):
        if not self._obj:
            return ""
        obj = self._obj
        if not obj.strip("(").startswith("/"):
            if "#" in obj:
                pos = obj.find("#")
                nodename = obj[:pos] or "*"
                nodeid = obj[pos + 1:]
                t = nodename + nodeid
                if not nodeid or '"' in t or "'" in t:
                    obj = "/" + obj
                else:
                    obj = "//%s[@id='%s']" % (nodename, nodeid)
            else:
                obj = "/" + obj
            self._obj = obj

        return str(obj)

    def __repr__(self):
        return repr(str(self))

    @property
    def Axis(self):
        axis = None
        pos = self._censor().strip("/").find("::")
        if pos > 0:
            axis = self.strip("/")[:pos]
        return axis

    @property
    def Nodetest(self):
        p1 = self._censor().strip("/").rfind("::")
        if p1 < 0:
            p1 = self._censor().strip("/").rfind("/")
        if p1 < 0:
            p1 = 0
        p2 = self._censor().strip("/").find("[", p1)
        if p2 < 0:
            p2 = len(self._censor().strip("/"))
        nodetest = self._censor().strip("/")[p1 + 1:p2]
        # print nodetest
        return nodetest

    def break_steps(self):
        ret = []
        p1 = 0
        p2 = 0
        p3 = 0
        s = str(self)
        t = self._censor()
        while True:
            if s[p1] == "(":
                p2 = p1
                while p2 >= 0:
                    p2 = t.find(")", p2 + 1)
                    if t.count("(", p1 + 1, p2) == t.count(")", p1 + 1, p2):
                        break
            else:
                p2 = p1 + 2
            p3 = t.find("/", p2)
            if p3 < 0:
                break
            if s[p3 - 1] == "(":
                p3 = p3 - 1
            ret.append(XPath(s[p1:p3]))
            p1 = p3
        ret.append(XPath(s[p1:]))
        return ret

    def break_frames(self):
        steps = self.break_steps()
        ret = []
        frame = ""
        for step in steps:
            frame += step
            if step.Nodetest.lower() in ["iframe", "frame"]:
                ret.append(XPath(frame))
                frame = ""
        if frame:
            ret.append(XPath(frame))
        return ret
    
    def _censor(self):
        if self._censored is not None:
            return self._censored
        ret = str(self)
        p3 = 0
        while p3 >= 0:
            p1 = ret.find("'", p3 + 1)
            p2 = ret.find('"', p3 + 1)
            if p1 < 0 and p2 < 0:
                break
            if p2 < 0 or 0 <= p1 < p2:
                p3 = ret.find("'", p1 + 1)
                if p3 < 0:
                    break
                ret = ret[:p1 + 1] + '"' * (p3 - p1 - 1) + ret[p3:]
            elif p1 < 0 or 0 <= p2 < p1:
                p3 = ret.find('"', p2 + 1)
                if p3 < 0:
                    break
                ret = ret[:p2 + 1] + "'" * (p3 - p2 - 1) + ret[p3:]
        self._censored = ret
        return ret

def set_logger(logger):
    '''set qt4w default logger
    '''
    import qt4w.util
    util.logger = logger



# -*- coding:UTF-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#

import os
from setuptools import setup, find_packages
  
BASE_DIR = os.path.realpath(os.path.dirname(__file__))
  
def generate_version():
    version = "2.0.0"
    if os.path.isfile(os.path.join(BASE_DIR, "version.txt")):
        with open("version.txt", "r") as fd:
            content = fd.read().strip()
            if content:
                version = content
    return version
  
def parse_requirements():
    reqs = []
    if os.path.isfile(os.path.join(BASE_DIR, "requirements.txt")):
        with open(os.path.join(BASE_DIR, "requirements.txt"), 'r') as fd:
            for line in fd.readlines():
                line = line.strip()
                if line:
                    reqs.append(line)
    return reqs

def list_data_files(packages):
    '''遍历package中的文件并查找所有资源文件
    
    :returns: list
    '''
    data_files = []
    for package in packages:
        file_path = os.path.join(BASE_DIR, *package.split('.'))
        if os.path.isdir(file_path):
            for dirpath, dirnames, filenames in os.walk(file_path):
                dirname = '/'.join(os.path.relpath(dirpath, BASE_DIR).split(os.path.sep))
                child_data_files = []
                for filename in filenames:
                    if not filename.endswith('.py') and not filename.endswith('.pyc'):
                        child_data_files.append(dirname+'/'+filename)
                if child_data_files:
                    data_files.append((dirname, child_data_files))
        else:
            dirname = '/'.join(package.split('.')[0:-1])    
            data_files.append(dirname+'/'+package.split('.')[-1])
    return data_files 

def get_description():
    with open(os.path.join(BASE_DIR, "README.md"), "rb") as fh:
        return fh.read().decode('utf8')
     
if __name__ == "__main__":  
    setup(
        zip_safe=False,
        version=generate_version(),
        name="qt4w",
        cmdclass={},
        packages=find_packages(exclude=("test", "test.*",)),
        include_package_data=True,
        description="QTA driver for Web",
        long_description=get_description(),
        long_description_content_type="text/markdown",
        author="Tencent",
        license="Copyright(c)2010-2018 Tencent All Rights Reserved. ",
        install_requires=parse_requirements(),  
        classifiers=[
          "Programming Language :: Python :: 2.7",
          "Programming Language :: Python :: 3.5",
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: 3.7",
          "Operating System :: OS Independent",
        ],
        url="https://github.com/qtacore/QT4W",
        project_urls={"QT4W Documentation":"https://qt4w.readthedocs.io/zh_CN/latest/",},
    )



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this
# file except in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''IWebView接口
'''



class IWebView(object):
    '''IWebView接口
    '''
    
    @property
    def webdriver_class(self):
        '''WebView对应的WebDriver类
        '''
        raise NotImplementedError
        
    @property
    def rect(self):
        '''WebView控件的坐标信息
        '''
        raise NotImplementedError
    
    @property
    def visible_rect(self):
        '''WebView控件可见区域的坐标信息
        '''
        return self.rect
    
    def eval_script(self, frame_xpaths, script):
        '''在指定frame中执行JavaScript，并返回执行结果
        
        :param frame_xpaths: frame元素的XPATH路径，如果是顶层页面，则传入“[]”
        :type frame_xpaths:  list
        :param script:       要执行的JavaScript语句
        :type script:        string
        '''
        raise NotImplementedError
    
    def screenshot(self):
        '''当前WebView的截图
        :return: PIL.Image
        '''
        raise NotImplementedError
        
    def click(self, x_offset, y_offset):
        '''点击WebView中的某个坐标
        
        :param x_offset: 与WebView左上角的横向偏移量
        :type x_offset:  int/float
        :param y_offset: 与WebView左上角的纵向偏移量
        :type y_offset:  int/float
        '''
        raise NotImplementedError
    
    def send_keys(self, text):
        '''发送可见字符按键
        
        :param text: 要输入的文本
        :type  text: string
        '''
        raise NotImplementedError
        
    def long_click(self, x_offset, y_offset, duration=1):
        '''长按WebView中的某个坐标
        
        :param x_offset: 与WebView左上角的横向偏移量
        :type x_offset:  int/float
        :param y_offset: 与WebView左上角的纵向偏移量
        :type y_offset:  int/float
        :param duration: 按住的持续时间
        :type duration:  int/float
        '''
        raise NotImplementedError
    
    def right_click(self, x_offset, y_offset):
        '''右键点击WebView中的某个坐标
        
        :param x_offset: 与WebView左上角的横向偏移量
        :type x_offset:  int/float
        :param y_offset: 与WebView左上角的纵向偏移量
        :type y_offset:  int/float
        '''
        raise NotImplementedError
    
    def double_click(self, x_offset, y_offset):
        '''双击WebView中的某个坐标
        
        :param x_offset: 与WebView左上角的横向偏移量
        :type x_offset:  int/float
        :param y_offset: 与WebView左上角的纵向偏移量
        :type y_offset:  int/float
        '''
        raise NotImplementedError
    
    def drag(self, x1, y1, x2, y2):
        '''从(x1, y1)点拖动到(x2, y2)点
        
        :param x1: 起点横坐标
        :type x1:  int/float
        :param y1: 起点纵坐标
        :type y1:  int/float
        :param x2: 终点横坐标
        :type x2:  int/float
        :param y2: 终点纵坐标
        :type y2:  int/float
        '''
        raise NotImplementedError
    
    def hover(self, x_offset, y_offset):
        '''
        
        :param x_offset: 与WebView左上角的横向偏移量
        :type x_offset:  int/float
        :param y_offset: 与WebView左上角的纵向偏移量
        :type y_offset:  int/float
        '''
        raise NotImplementedError
    
    def scroll(self, backward=True):
        '''
        
        :param backward: 是否向后滚动，默认为True
        :type  backward: bool
        '''
        raise NotImplementedError
     
    def upload_file(self, file_path):
        '''上传文件
        
        :param file_path: 文件路径
        :type  file_path: str
        '''
        raise NotImplementedError
     
# class WebViewBase(IWebView):
#     '''WebView基类，实现一些公共方法
#     '''
#     
#     def _handle_result(self, result, frame_xpaths):
#         '''处理执行JavaScript的结果
#         
#         :param result: 要处理的数据
#         :type  result: string
#         :param frame_xpaths: 执行js所在frame的xpath
#         :type  frame_xpaths: list
#         '''
#         from qt4w.util import JavaScriptError
#         if result[0] == 'S': return result[1:]
#         elif result[0] == 'E':
#             raise JavaScriptError(frame_xpaths, result[1:])
#         else:
#             raise ValueError('执行JavaScript返回结果错误：%r' % result)
            
